// bmp2nes.cpp

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define  MODE_LIST 0x0001
#define  FILE_IN   0x0100
#define  FILE_OUT  0x0200

#define  POS_MAX   256

typedef struct {
	short xpos;
	short ypos;
} pos_t;

typedef struct {
	int x;
	int y;
	int width;
	int height;
	int xstep;
	int ystep;
	int count;
	int append;
	int unified;
	int uni_cnt;
	int uni_num;
	char *uni_dict;
	char *list_name;
	int pos_num;
	pos_t *pos_list;
} param_t;

int  ParseArg(int argc, char *argv[], int *mode, char *inname, char *outname, param_t *param);
void Convert(FILE *ifp, FILE *ofp, param_t *param);
void Convert2(FILE *ifp, FILE *ofp, param_t *param);
void ReadPosList(FILE *pfp, param_t *param);

int main(int argc, char *argv[]) {
	FILE *ifp = NULL;
	FILE *ofp = NULL;
	FILE *pfp = NULL;
	char inname[256];
	char outname[256];
	pos_t pos_list[POS_MAX];
	param_t param;
	char *dict = NULL;
	int mode;
	int r;

	// 
	ParseArg(argc, argv, &mode, inname, outname, &param);

	// m
	dict = new char[16*4096];
	param.uni_dict = dict;

	param.pos_list = pos_list;

	// t@CJ
	if ((mode & (FILE_IN|FILE_OUT)) != (FILE_IN|FILE_OUT)) {
		goto bye;
	}

	ifp = fopen(inname, "rb");
	if (ifp == NULL) {
		printf("%s ̃t@CI[vɎs܂B\n", inname);
		goto bye;
	}

	if (param.unified && param.append) {
		ofp = fopen(outname, "rb");
		if (ofp == NULL) {
			printf("%s ̃t@CI[vɎs܂B\n", outname);
			goto bye;
		}
		r = fread(dict, 16*4096, sizeof(char), ofp);
		param.uni_cnt = r / 16;
		fclose(ofp);
		ofp = NULL;
	}

	if (!param.unified && param.append) {
		ofp = fopen(outname, "ab");
	} else {
		ofp = fopen(outname, "wb");
	}
	if (ofp == NULL) {
		printf("%s ̃t@CI[vɎs܂B\n", outname);
		goto bye;
	}

	if (mode & MODE_LIST) {
		pfp = fopen(param.list_name, "r");
		if (pfp == NULL) {
			printf("%s ̃t@CI[vɎs܂B\n", param.list_name);
			goto bye;
		}
	}

	if (mode & MODE_LIST) {
		ReadPosList(pfp, &param);
		Convert2(ifp, ofp, &param);
	} else {
		Convert(ifp, ofp, &param);
	}

bye:
	if (ifp != NULL) fclose(ifp);
	if (ofp != NULL) fclose(ofp);
	if (pfp != NULL) fclose(pfp);
	delete[] dict;
}

int ParseArg(int argc, char *argv[], int *mode, char *inname, char *outname, param_t *param) {
	char *ep;
	int phase = 0;
	int i, n;

	memset(param, 0, sizeof(param_t));

	*mode = 0;
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-bg2") == 0) {
			param->xstep = 1;
			param->ystep = 1;
		} else if (strcmp(argv[i], "-bg2z") == 0) {
			param->xstep = 2;
			param->ystep = 2;
		} else if (strcmp(argv[i], "-sp2") == 0) {
			param->xstep = 1;
			param->ystep = 2;
		} else if (strcmp(argv[i], "-pl") == 0) {
			i++;
			if (i < argc) {
				param->list_name = argv[i];
			}
			*mode |= MODE_LIST;
		} else if (strcmp(argv[i], "-x") == 0) {
			i++;
			if (i < argc) {
				param->x = strtol(argv[i], NULL, 0);
			}
		} else if (strcmp(argv[i], "-y") == 0) {
			i++;
			if (i < argc) {
				param->y = strtol(argv[i], NULL, 0);
			}
		} else if (strcmp(argv[i], "-width") == 0) {
			i++;
			if (i < argc) {
				param->width = strtol(argv[i], NULL, 0);
			}
		} else if (strcmp(argv[i], "-height") == 0) {
			i++;
			if (i < argc) {
				param->height = strtol(argv[i], NULL, 0);
			}
		} else if (strcmp(argv[i], "-xstep") == 0) {
			i++;
			if (i < argc) {
				param->xstep = strtol(argv[i], NULL, 0);
			}
		} else if (strcmp(argv[i], "-ystep") == 0) {
			i++;
			if (i < argc) {
				param->ystep = strtol(argv[i], NULL, 0);
			}
		} else if (strcmp(argv[i], "-n") == 0) {
			i++;
			if (i < argc) {
				param->count = strtol(argv[i], NULL, 0);
			}
		} else if (strcmp(argv[i], "-a") == 0) {
			param->append = 1;
		} else if (strcmp(argv[i], "-uni") == 0) {
			param->unified = 1;
			if (i+1 < argc) {
				n = strtol(argv[i+1], &ep, 0);
				if ((ep != argv[i+1]) && (*ep == '\0')) {
					i++;
					param->uni_num = n;
				}
			}
		} else {
			if (phase == 0) {
				strcpy(inname, argv[i]);
				*mode |= FILE_IN;
			} else if (phase == 1) {
				strcpy(outname, argv[i]);
				*mode |= FILE_OUT;
			}
			phase++;
		}
	}

	if (param->xstep < 1) param->xstep = 0;
	if (param->ystep < 1) param->ystep = 0;
}

void Convert(FILE *ifp, FILE *ofp, param_t *param) {
	char  head[64];
	char *src = NULL;
	char *dst = NULL;
	char *p1, *p2, *p3, *p4;
	int   offset;
	int   width;
	int   height;
	int   width2;
	int   height2;
	int   xstep;
	int   ystep;
	int   pitch;
	int   count;
	int   data;
	int   i, j, x, y, w, h;

	// wb_[̓ǂݍ
	fread(head, 54, sizeof(char), ifp);
	width   = (head[18] & 255);
	width  |= (head[19] & 255) <<  8;
	width  |= (head[20] & 255) << 16;
	width  |= (head[21] & 255) << 24;
	height  = (head[22] & 255);
	height |= (head[23] & 255) <<  8;
	height |= (head[24] & 255) << 16;
	height |= (head[25] & 255) << 24;
	pitch   = (width+3) & -4;

	xstep = param->xstep * 8;
	ystep = param->ystep * 8;
	width2  = param->width;
	height2 = param->height;
	if (width2  == 0) width2  = width;
	if (height2 == 0) height2 = height;

	// obt@̊m
	src = new char[pitch * ystep];
	dst = new char[16];

	// sNZϊ
	count = 0;
	offset = 54+1024;
	offset += pitch * (height - param->y);
	for (h = 0; h < height2 - (ystep-1); h += ystep) {
		offset -= pitch * ystep;
		fseek(ifp, offset, SEEK_SET);
		fread(src, pitch*ystep, sizeof(char), ifp);
		p1 = src + pitch*ystep + param->x;
		for (w = 0; w < width2 - (xstep-1); w += xstep) {
			p2 = p1;
			for (y = 0; y < ystep; y += 8) {
				p3 = p2;
				for (x = 0; x < xstep; x += 8) {
					p4 = p3;
					for (j = 0; j < 8; j++) {
						p4 -= pitch;
						data = 0;
						for (i = 0; i < 8; i++) {
							data |= ((0x00060500 >> (p4[i] & 3)) & 0x8080) >> i;
						}
						dst[j  ] = (char)(data     );
						dst[j+8] = (char)(data >> 8);
					}
					if (param->unified) {
						for (j = 0; j < param->uni_cnt; j++) {
							if (memcmp(&(param->uni_dict[j*16]), dst, 16) == 0) break;
						}
						if (j >= param->uni_cnt) {
							memcpy(&(param->uni_dict[param->uni_cnt*16]), dst, 16);
							param->uni_cnt++;
						}
					} else {
						fwrite(dst, 16, sizeof(char), ofp);
					}
					count++;
					if ((param->count > 0) && (count >= param->count)) {
						goto break_up;
					}
					p3 += 8;
				}
				p2 -= pitch * 8;
			}
			p1 += xstep;
		}
	}
break_up:

	if (param->unified) {
		if ((param->uni_num) && (param->uni_cnt < param->uni_num)) {
			memset(&(param->uni_dict[param->uni_cnt*16]), 0xff, (param->uni_num-param->uni_cnt)*16);
			param->uni_cnt = param->uni_num;
		}
		fwrite(param->uni_dict, param->uni_cnt*16, sizeof(char), ofp);
	}

	// obt@̊J
	delete src;
	delete dst;
}

void Convert2(FILE *ifp, FILE *ofp, param_t *param) {
	char  head[64];
	char *src = NULL;
	char *dst = NULL;
	char *p1;
	int   offset;
	int   width;
	int   height;
	int   xstep;
	int   ystep;
	int   pitch;
	int   data;
	int   i, j, x, y, k;

	// wb_[̓ǂݍ
	fread(head, 54, sizeof(char), ifp);
	width   = (head[18] & 255);
	width  |= (head[19] & 255) <<  8;
	width  |= (head[20] & 255) << 16;
	width  |= (head[21] & 255) << 24;
	height  = (head[22] & 255);
	height |= (head[23] & 255) <<  8;
	height |= (head[24] & 255) << 16;
	height |= (head[25] & 255) << 24;
	pitch   = (width+3) & -4;

	xstep = param->xstep * 8;
	ystep = param->ystep * 8;

	// obt@̊m
	src = new char[pitch * ystep];
	dst = new char[16];

	// sNZϊ
	for (k = 0; k < param->pos_num; k++) {
		x = param->pos_list[k].xpos;
		y = param->pos_list[k].ypos;

		offset = 54+1024;
		offset += pitch * (height - (param->y + y));
		offset -= pitch * 8;
		fseek(ifp, offset, SEEK_SET);
		fread(src, pitch*ystep, sizeof(char), ifp);
		p1 = src + pitch*ystep + param->x + x;
		for (j = 0; j < 8; j++) {
			p1 -= pitch;
			data = 0;
			for (i = 0; i < 8; i++) {
				data |= ((0x00060500 >> (p1[i] & 3)) & 0x8080) >> i;
			}
			dst[j  ] = (char)(data     );
			dst[j+8] = (char)(data >> 8);
		}
		if (param->unified) {
			for (j = 0; j < param->uni_cnt; j++) {
				if (memcmp(&(param->uni_dict[j*16]), dst, 16) == 0) break;
			}
			if (j >= param->uni_cnt) {
				memcpy(&(param->uni_dict[param->uni_cnt*16]), dst, 16);
				param->uni_cnt++;
			}
		} else {
			fwrite(dst, 16, sizeof(char), ofp);
		}
	}

	if (param->unified) {
		if ((param->uni_num) && (param->uni_cnt < param->uni_num)) {
			memset(&(param->uni_dict[param->uni_cnt*16]), 0xff, (param->uni_num-param->uni_cnt)*16);
			param->uni_cnt = param->uni_num;
		}
		fwrite(param->uni_dict, param->uni_cnt*16, sizeof(char), ofp);
	}

	// obt@̊J
	delete src;
	delete dst;
}

void ReadPosList(FILE *pfp, param_t *param) {
	char line[1024];
	char *p, *q;

	param->pos_num = 0;
	while (!feof(pfp)) {
		p = fgets(line, sizeof(line), pfp);
		if (p == NULL) break;

		while (*p) {
			while (isspace(*p)) p++;
			if (*p == '(') p++;

			while (isspace(*p)) p++;
			param->pos_list[param->pos_num].xpos = strtol(p, &q, 0);
			if (q == p) break;
			p = q;

			while (isspace(*p)) p++;
			if (*p == ',') p++;

			while (isspace(*p)) p++;
			param->pos_list[param->pos_num].ypos = strtol(p, &q, 0);
			if (q == p) break;
			p = q;

			while (isspace(*p)) p++;
			if (*p == ')') p++;

			while (isspace(*p)) p++;
			param->pos_num++;
		}
	}
}

